#include <output/WriterBase.hpp>

#include <mesh/IMesh.hpp>
#include <output/OutputNamedItemValueSet.hpp>
#include <scheme/DiscreteFunctionP0.hpp>
#include <scheme/DiscreteFunctionP0Vector.hpp>
#include <scheme/IDiscreteFunction.hpp>
#include <scheme/IDiscreteFunctionDescriptor.hpp>
#include <utils/Exceptions.hpp>

template <typename DiscreteFunctionType>
void
WriterBase::_register(const std::string& name,
                      const DiscreteFunctionType& discrete_function,
                      OutputNamedItemDataSet& named_item_data_set)
{
  if constexpr (DiscreteFunctionType::handled_data_type == IDiscreteFunction::HandledItemDataType::value) {
    named_item_data_set.add(NamedItemData{name, discrete_function.cellValues()});
  } else {
    named_item_data_set.add(NamedItemData{name, discrete_function.cellArrays()});
  }
}

template <size_t Dimension, template <size_t DimensionT, typename DataTypeT> typename DiscreteFunctionType>
void
WriterBase::_register(const std::string& name,
                      const IDiscreteFunction& i_discrete_function,
                      OutputNamedItemDataSet& named_item_data_set)
{
  const ASTNodeDataType& data_type = i_discrete_function.dataType();
  switch (data_type) {
  case ASTNodeDataType::bool_t: {
    _register(name, dynamic_cast<const DiscreteFunctionType<Dimension, bool>&>(i_discrete_function),
              named_item_data_set);
    break;
  }
  case ASTNodeDataType::unsigned_int_t: {
    _register(name, dynamic_cast<const DiscreteFunctionType<Dimension, uint64_t>&>(i_discrete_function),
              named_item_data_set);
    break;
  }
  case ASTNodeDataType::int_t: {
    _register(name, dynamic_cast<const DiscreteFunctionType<Dimension, int64_t>&>(i_discrete_function),
              named_item_data_set);
    break;
  }
  case ASTNodeDataType::double_t: {
    _register(name, dynamic_cast<const DiscreteFunctionType<Dimension, double>&>(i_discrete_function),
              named_item_data_set);
    break;
  }
  case ASTNodeDataType::vector_t: {
    if constexpr (DiscreteFunctionType<Dimension, double>::handled_data_type ==
                  IDiscreteFunction::HandledItemDataType::vector) {
      throw UnexpectedError("invalid data type for vector data");
    } else {
      switch (data_type.dimension()) {
      case 1: {
        _register(name,
                  dynamic_cast<const DiscreteFunctionType<Dimension, TinyVector<1, double>>&>(i_discrete_function),
                  named_item_data_set);
        break;
      }
      case 2: {
        _register(name,
                  dynamic_cast<const DiscreteFunctionType<Dimension, TinyVector<2, double>>&>(i_discrete_function),
                  named_item_data_set);
        break;
      }
      case 3: {
        _register(name,
                  dynamic_cast<const DiscreteFunctionType<Dimension, TinyVector<3, double>>&>(i_discrete_function),
                  named_item_data_set);
        break;
      }
      default: {
        throw UnexpectedError("invalid vector dimension");
      }
      }
    }
    break;
  }
  case ASTNodeDataType::matrix_t: {
    if constexpr (DiscreteFunctionType<Dimension, double>::handled_data_type ==
                  IDiscreteFunction::HandledItemDataType::vector) {
      throw UnexpectedError("invalid data type for vector data");
    } else {
      Assert(data_type.numberOfRows() == data_type.numberOfColumns(), "invalid matrix dimensions");
      switch (data_type.numberOfRows()) {
      case 1: {
        _register(name,
                  dynamic_cast<const DiscreteFunctionType<Dimension, TinyMatrix<1, 1, double>>&>(i_discrete_function),
                  named_item_data_set);
        break;
      }
      case 2: {
        _register(name,
                  dynamic_cast<const DiscreteFunctionType<Dimension, TinyMatrix<2, 2, double>>&>(i_discrete_function),
                  named_item_data_set);
        break;
      }
      case 3: {
        _register(name,
                  dynamic_cast<const DiscreteFunctionType<Dimension, TinyMatrix<3, 3, double>>&>(i_discrete_function),
                  named_item_data_set);
        break;
      }
      default: {
        throw UnexpectedError("invalid matrix dimension");
      }
      }
    }
    break;
  }
  default: {
    throw UnexpectedError("invalid data type " + dataTypeName(data_type));
  }
  }
}

template <template <size_t Dimension, typename DataType> typename DiscreteFunctionType>
void
WriterBase::_register(const NamedDiscreteFunction& named_discrete_function, OutputNamedItemDataSet& named_item_data_set)
{
  const IDiscreteFunction& i_discrete_function = *named_discrete_function.discreteFunction();
  const std::string& name                      = named_discrete_function.name();
  switch (i_discrete_function.mesh()->dimension()) {
  case 1: {
    _register<1, DiscreteFunctionType>(name, i_discrete_function, named_item_data_set);
    break;
  }
  case 2: {
    _register<2, DiscreteFunctionType>(name, i_discrete_function, named_item_data_set);
    break;
  }
  case 3: {
    _register<3, DiscreteFunctionType>(name, i_discrete_function, named_item_data_set);
    break;
  }
  default: {
    throw UnexpectedError("invalid mesh dimension");
  }
  }
}

std::shared_ptr<const IMesh>
WriterBase::_getMesh(
  const std::vector<std::shared_ptr<const NamedDiscreteFunction>>& named_discrete_function_list) const
{
  Assert(named_discrete_function_list.size() > 0);

  std::shared_ptr mesh = named_discrete_function_list[0]->discreteFunction()->mesh();
  for (size_t i = 1; i < named_discrete_function_list.size(); ++i) {
    if (mesh != named_discrete_function_list[i]->discreteFunction()->mesh()) {
      std::ostringstream error_msg;
      error_msg << "discrete functions must be defined on the same mesh!\n"
                << rang::fgB::yellow << "note:" << rang::style::reset << " cannot write " << rang::fgB::blue
                << named_discrete_function_list[0]->name() << rang::style::reset << " and " << rang::fgB::blue
                << named_discrete_function_list[i]->name() << rang::style::reset << " in the same file.";
      throw NormalError(error_msg.str());
    }
  }

  return mesh;
}

OutputNamedItemDataSet
WriterBase::_getOutputNamedItemDataSet(
  const std::vector<std::shared_ptr<const NamedDiscreteFunction>>& named_discrete_function_list) const
{
  OutputNamedItemDataSet named_item_data_set;

  for (auto& named_discrete_function : named_discrete_function_list) {
    const IDiscreteFunction& i_discrete_function = *named_discrete_function->discreteFunction();

    switch (i_discrete_function.descriptor().type()) {
    case DiscreteFunctionType::P0: {
      WriterBase::_register<DiscreteFunctionP0>(*named_discrete_function, named_item_data_set);
      break;
    }
    case DiscreteFunctionType::P0Vector: {
      WriterBase::_register<DiscreteFunctionP0Vector>(*named_discrete_function, named_item_data_set);
      break;
    }
    default: {
      std::ostringstream error_msg;
      error_msg << "the type of discrete function of " << rang::fgB::blue << named_discrete_function->name()
                << rang::style::reset << " is not supported";
      throw NormalError(error_msg.str());
    }
    }
  }

  return named_item_data_set;
}

WriterBase::WriterBase(const std::string& base_filename, const double& time_period)
  : m_base_filename{base_filename}, m_period_manager(time_period)
{}

WriterBase::WriterBase(const std::string& base_filename) : m_base_filename{base_filename} {}

void
WriterBase::writeIfNeeded(const std::vector<std::shared_ptr<const NamedDiscreteFunction>>& named_discrete_function_list,
                          double time) const
{
  if (m_period_manager.has_value()) {
    const double last_time = m_period_manager->getLastTime();
    if (time == last_time)
      return;   // output already performed

    if (time >= m_period_manager->nextTime()) {
      this->_writeAtTime(named_discrete_function_list, time);
      m_period_manager->setSaveTime(time);
    }
  } else {
    throw NormalError("this writer does not allow time value");
  }
}

void
WriterBase::writeForced(const std::vector<std::shared_ptr<const NamedDiscreteFunction>>& named_discrete_function_list,
                        double time) const
{
  if (m_period_manager.has_value()) {
    if (time == m_period_manager->getLastTime())
      return;   // output already performed

    this->_writeAtTime(named_discrete_function_list, time);
    m_period_manager->setSaveTime(time);
  } else {
    throw NormalError("this writer does not allow time value");
  }
}

void
WriterBase::write(const std::vector<std::shared_ptr<const NamedDiscreteFunction>>& named_discrete_function_list) const
{
  if (m_period_manager.has_value()) {
    throw NormalError("this writer requires time value");
  } else {
    this->_write(named_discrete_function_list);
  }
}

void
WriterBase::writeMesh(const std::shared_ptr<const IMesh>& mesh) const
{
  if (m_period_manager.has_value()) {
    throw NormalError("write_mesh requires a writer without time period");
  } else {
    this->_writeMesh(mesh);
  }
}
